from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader


class CLR1:
    """A class that represents a CLR1 block in a MSBP file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBP-File-Format#clr1-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.colors: list[dict] = []

    def read(self, reader: Reader) -> None:
        """Reads the CLR1 block from a stream.

        :param `reader`: A Reader object."""
        self.block.read_header(reader)

        color_count = reader.read_uint32()
        # Read the RGBA colors
        for _ in range(color_count):
            color = {}
            color["r"] = reader.read_uint8()
            color["g"] = reader.read_uint8()
            color["b"] = reader.read_uint8()
            color["a"] = reader.read_uint8()
            self.colors.append(color)

        self.block.seek_to_end(reader)



from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer


class LMS_Block:
    """A class that represents a block in a LMS file."""

    def __init__(self):
        self.magic: str = None
        self.size: int = 0
        self.data_start: int = 0

    def read_header(self, reader: Reader) -> None:
        """Reads the block header from a stream

        `reader`: A Reader object."""
        self.magic = reader.read_string_len(4)
        self.size = reader.read_uint32()
        reader.skip(8)
        self.data_start = reader.tell()

    def get_item_offsets(self, reader: Reader, item_count: int) -> list[int]:
        """Returns the absolute offsets to items in a block.

        :param `reader`: A Reader object.
        :param `item_count`: Amount of items."""
        return [reader.read_uint32() + self.data_start for _ in range(item_count)]

    def seek_to_end(self, stream: Reader | Writer) -> None:
        """Seeks to the end of a block passed the AB padding

        `stream`: A Writer or Reader object."""
        stream.seek(self.data_start)
        end = self.size + 16 - self.size % 16
        stream.seek(end, 1)

    def write_initial_data(self, writer: Writer, item_count: int | None = None) -> None:
        """Sets and writes the initial data for a block.
        
        :param `writer`: a Writer object.
        :param `magic`: the magic name of the block."""
        writer.write_string(self.magic)
        writer.write_uint32(0)
        writer.write_bytes(b"\x00" * 8)
        self.data_start = writer.tell()

        if item_count:
            writer.write_uint32(item_count)

    def write_end_data(self, writer: Writer):
        """Writes the ab padding, size, and seeks to the end of a block.

        `writer`: A Writer object."""
        offset = writer.tell()

        # 0xAB padding
        remainder = 16 - self.size % 16
        if remainder == 16:
            return 0

        writer.write_bytes(b"\xAB" * remainder)

        # Size
        writer.seek(self.data_start - 12)
        writer.write_uint32(self.size)
        writer.seek(offset)

        if remainder == 0:
            return

        writer.write_bytes(b"\xAB" * remainder)
        self.seek_to_end(writer)#



from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer


class LMS_HashTable:
    """A class that represents a generic hash table block.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/LMS-File-Format#hash-tables"""

    def __init__(self):
        self.block = LMS_Block()
        self.labels: dict[int:str] = {}

    def get_index_by_label(self, label: str) -> None:
        """Returns the index of a label given its name

        :param `label`: The label to find the index of."""
        for index in self.labels:
            if self.labels[index] == label:
                return index

    def add_label(self, label: str) -> None:
        """Adds a label to the hash tabel block.

        param `label`: The label to add."""
        for index in self.labels:
            if self.labels[index] == label:
                raise ValueError("This label is already in the block.")

        index = len(self.labels)
        self.labels[index] = label

    def delete_label(self, label: str) -> None:
        """Deletes a label in the hash table block

        :param `label`: The label to delete."""
        found = False
        for index in self.labels:
            if self.labels[index] == label:
                found = True

        if not found:
            raise ValueError("This label does not exist in the block.")

        index = self.get_index_by_label(label)

        del self.labels[index]

        # Resize the indexes
        for i in range(index, len(self.labels)):
            self.labels[i] = self.labels[i + 1]

        # Remove the last element to avoid duplication
        self.labels.pop(len(self.labels) - 1, None)

    def edit_label(self, label: str, new_labeL: str) -> None:
        """Edits a label in the hash table block.

        :param `label`: The existing label to edit.
        :param `new_label`: The new label.
        """
        index = self.get_index_by_label(label)
        self.labels[index] = new_labeL

    def read(self, reader: Reader) -> None:
        """Reads the hash table block from a stream.

        :param `reader`: A Reader object."""
        self.block.read_header(reader)
        slot_count = reader.read_uint32()

        for _ in range(slot_count):
            label_count = reader.read_uint32()
            offset = reader.read_uint32()
            end = reader.tell()
            reader.seek(self.block.data_start)
            reader.seek(offset, 1)

            for _ in range(label_count):
                length = reader.read_uint8()
                label = reader.read_string_len(length)
                item_index = reader.read_uint32()
                self.labels[item_index] = label

            reader.seek(end)

        sorted_labels = {}

        for index in sorted(self.labels):
            sorted_labels[index] = self.labels[index]

        self.labels = sorted_labels

    def calc_hash(self, label, num_slots) -> int:
        """Returns the hash of a label.

        :param `label`: The label.
        :param `num_slots`: The amount of hash table slots.

        https://github.com/kinnay/Nintendo-File-Formats/wiki/LMS-File-Format#hash-tables
        """
        hash = 0
        for char in label:
            hash = hash * 0x492 + ord(char)
        return (hash & 0xFFFFFFFF) % num_slots

    def write(self, writer: Writer, slot_count: int) -> None:
        """Writes the hash table block to a stream.

        :param `writer`: A Writer object.
        :param `slot_count`: The hash table slot count. 101 for MSBT, 29 for MSBP and 59 for MSBF.
        """
        self.block.write_initial_data(writer, slot_count)
      
        hash_slots = {}
        # Add each label to each hash slot
        for index in self.labels:
            label = self.labels[index]
            hash = self.calc_hash(label, slot_count)
            if hash not in hash_slots:
                hash_slots[hash] = [label]
            else:
                hash_slots[hash].append(label)
        size = 0
        label_offsets = slot_count * 8 + 4
        size += label_offsets

        hash_slots = dict(sorted(hash_slots.items(), key=lambda x: x[0]))

        # Write the slots
        for i in range(slot_count):
            if i in hash_slots:
                writer.write_uint32(len(hash_slots[i]))
                writer.write_uint32(label_offsets)

                for label in hash_slots[i]:
                    label_offsets += len(label) + 5
            else:
                writer.write_uint32(0)
                writer.write_uint32(label_offsets)

        # Write the labels
        for key in hash_slots:
            labels = hash_slots[key]
            for label in labels:
                label_index = self.get_index_by_label(label)
                writer.write_uint8(len(label))
                writer.write_string(label)
                writer.write_uint32(label_index)
                size += 5 + len(label)

        self.block.size = size
        self.block.write_end_data(writer)



from LMS.Common.LMS_Binary import LMS_Binary
from LMS.Common.LMS_HashTable import LMS_HashTable
from LMS.Common.LMS_Enum import LMS_BinaryTypes
from LMS.Stream.Reader import Reader

from LMS.Project.BaseBlock import ProjectBlock
from LMS.Project.Structure import TagGroup, Tag, TagParameter, AttributeStructure

from LMS.Project.CLR1 import CLR1
from LMS.Project.CTI1 import CTI1

from LMS.Project.ALI2 import ALI2

from LMS.Project.TGL2 import TGL2

from LMS.Project.SYL3 import SYL3


class MSBP:
    """A class that represents a Message Studio Binary Project file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBP-File-Format"""

    def __init__(self):
        self.binary: LMS_Binary = LMS_Binary()
        self.CLR1: CLR1 = CLR1()
        self.CLB1: LMS_HashTable = LMS_HashTable()

        self.ATI2 = ProjectBlock(AttributeStructure)
        self.ALB1: LMS_HashTable = LMS_HashTable()
        self.ALI2 = ALI2()

        self.TGG2 = ProjectBlock(TagGroup)
        self.TAG2 = ProjectBlock(Tag)
        self.TGP2 = ProjectBlock(TagParameter)
        self.TGL2 = TGL2()

        self.SYL3 = SYL3()
        self.SLB1: LMS_HashTable = LMS_HashTable()

        self.CTI1 = CTI1()

    def get_attribute_structure(self) -> dict[str, AttributeStructure]:
        """Returns a formatted dictionary of all the data in the ATI2, ALB1, ALI2 blocks."""
        if self.ALB1 is None:
            return {}

        structure = {self.ALB1.labels[i]: self.ATI2.data[i] for i in self.ALB1.labels}

        for label in structure:
            attribute: AttributeStructure = structure[label]
            if attribute.type is LMS_BinaryTypes.LIST_INDEX:
                attribute.list_items = self.ALI2.attribute_lists[attribute.list_index]

        return structure

    def get_tag_structure(self) -> list[TagGroup]:
        """Returns a formatted dictionary of all the data in the TGG2, TAG2, TGP2, AND TGL2 blocks."""
        structure: list[TagGroup] = []
        
        for group in self.TGG2.data:
            group.tags = [self.TAG2.data[i] for i in group.tag_indexes]

            for tag in group.tags:
                tag.parameters = [self.TGP2.data[i] for i in tag.parameter_indexes]

            structure.append(group)

        list_parameters: list[TagParameter] = [
            parameter
            for parameter in self.TGP2.data
            if parameter.type is LMS_BinaryTypes.LIST_INDEX
        ]

        for parameter in list_parameters:
            parameter.list_items = [self.TGL2.items[i] for i in parameter.list_indexes]

        return structure

    def read(self, reader: Reader) -> None:
        """Reads a MSBP file from a stream.

        :param `reader`: A Reader object."""
        self.binary.read_header(reader)

        clr1_valid, clr1_offset = self.binary.search_block_by_name(reader, "CLR1")
        clb1_valid, clb1_offset = self.binary.search_block_by_name(reader, "CLB1")
        ati2_valid, ati2_offset = self.binary.search_block_by_name(reader, "ATI2")
        alb1_valid, alb1_offset = self.binary.search_block_by_name(reader, "ALB1")
        ali2_valid, ali2_offset = self.binary.search_block_by_name(reader, "ALI2")
        tgg2_valid, tgg2_offset = self.binary.search_block_by_name(reader, "TGG2")
        tag2_valid, tag2_offset = self.binary.search_block_by_name(reader, "TAG2")
        tgp2_valid, tgp2_offset = self.binary.search_block_by_name(reader, "TGP2")
        tgl2_valid, tgl2_offset = self.binary.search_block_by_name(reader, "TGL2")
        syl3_valid, syl3_offset = self.binary.search_block_by_name(reader, "SYL3")
        slb1_valid, slb1_offset = self.binary.search_block_by_name(reader, "SLB1")
        cti1_valid, cti1_offset = self.binary.search_block_by_name(reader, "CTI1")

        # Read CLR1
        if clr1_valid:
            reader.seek(clr1_offset)
            self.CLR1.read(reader)

        # Read CLB1
        if clb1_valid:
            reader.seek(clb1_offset)
            self.CLB1.read(reader)

        # Read ATI2
        if ati2_valid:
            reader.seek(ati2_offset)
            self.ATI2.read(reader, contains_offsets=False)

        # Read ALB1
        if alb1_valid:
            reader.seek(alb1_offset)
            self.ALB1.read(reader)

        # Read ALI2
        if ali2_valid:
            reader.seek(ali2_offset)
            self.ALI2.read(reader)

        # Read TGG2
        if tgg2_valid:
            reader.seek(tgg2_offset)
            self.TGG2.read(reader, version_4=self.binary.revision == 4)

        # Read TAG2
        if tag2_valid:
            reader.seek(tag2_offset)
            self.TAG2.read(reader, uint32_count=False)

        # Read TGP2
        if tgp2_valid:
            reader.seek(tgp2_offset)
            self.TGP2.read(reader, uint32_count=False)

        # Read TGL2
        if tgl2_valid:
            reader.seek(tgl2_offset)
            self.TGL2.read(reader)

        # Read SYL3
        if syl3_valid:
            reader.seek(syl3_offset)
            self.SYL3.read(reader)

        # Read SLB1
        if slb1_valid:
            reader.seek(slb1_offset)
            self.SLB1.read(reader)

        # Read CTI1
        if cti1_valid:
            reader.seek(cti1_offset)
            self.CTI1.read(reader)



from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer
from LMS.Project.MSBP import MSBP

from LMS.Message.Tags import Tag_Utility


class TXT2:
    """A class that represents a TXT2 block in a MSBT file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBT-File-Format#txt2-block"""

    def __init__(self):
        from LMS.Message.MSBT import MSBT

        self.msbt: MSBT = None

        self.block: LMS_Block = LMS_Block()
        self.messages: list[str] = []

        self.errors: dict[str:dict] = {}

    def read(self, reader: Reader, preset, lbl1) -> None:
        """Reads the TXT2 block from a stream.

        :param `reader`: A Reader object.
        :param `msbp`: A MSBP object used for decoding attributes and tags
        """
        self.block.read_header(reader)
        message_count = reader.read_uint32()

        encoding = reader.get_utf16_encoding()
        tag_indicator = b"\x0E\x00" if reader.byte_order == "little" else b"\x00\x0E"

        offsets = self.block.get_item_offsets(reader, message_count)

        for i, offset in enumerate(offsets):
            if i < len(offsets) - 1:
                next_offset = offsets[i + 1]
            else:
                next_offset = self.block.data_start + self.block.size

            reader.seek(offset)

            message = b""
            while reader.tell() < next_offset:
                bytes = reader.read_bytes(2)
                if bytes == tag_indicator:
                    tag, error = Tag_Utility.read_tag(reader, preset)
                    
                    if error:
                        label = lbl1.labels[i]
                        self.errors[label]["read"].append(error)
                        message += tag.encode(encoding)
                    else:
                        message += tag.encode(encoding)
                else:
                    message += bytes 

            self.messages.append(message.decode(encoding))

        

        self.block.seek_to_end(reader)

    def write(self, writer: Writer, preset, lbl1) -> None:
        """Writes the TXT2 block to a stream.

        :param `writer`: A Writer object.
        :param `msbp`: a MSBP object used for writing decoded attributes and tags.
        :param `lbl1`: a LBL1 object used for preset errors.
        """
        message_count = len(self.messages)
        self.block.write_initial_data(writer, message_count)
        message_offset = message_count * 4 + 4

        tag_indicator = b"\x0E\x00" if writer.byte_order == "little" else b"\x00\x0E"

        encoded_messages = []
        for i, message in enumerate(self.messages):
            split_message = Tag_Utility.split_message_by_tag(message)
            # Use a writer object to simplfy the encoding process
            message_writer = Writer(b"", writer.byte_order)

            for part in split_message:
                if Tag_Utility.is_tag(part):
                    message_writer.write_bytes(tag_indicator)
                    error = Tag_Utility.write_tag(message_writer, part, preset)
                    
                    if error:
                        label = lbl1.labels[i]
                        self.errors[label]["write"].append(error)
                else:
                    message_writer.write_utf16_string(part)

            encoded_messages.append(message_writer.get_data())

        size = message_count * 4 + 4
        for message in encoded_messages:
            writer.write_uint32(message_offset)
            message_offset += len(message)
            size += len(message)

        for message in encoded_messages:
            writer.write_bytes(message)

        self.block.size = size
        self.block.write_end_data(writer)



from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer
from LMS.Common.LMS_Enum import LMS_BinaryTypes

class AttributeStructure:
    """A class that represents the structure of a specifc attribute."""
    def __init__(self):
        self.type: LMS_BinaryTypes = None
        self.offset = 0 
        self.list_index = 0

        self.list_items: list[str] = []

    def read(self, reader: Reader) -> None:
        self.type = LMS_BinaryTypes(reader.read_uint8())
        reader.skip(1)
        self.list_index = reader.read_uint16()
        self.offset = reader.read_uint32()

class TagGroup:
    """A class that represents a tag group."""
    def __init__(self):
        self.name: str = ""
        self.group_index: int | None = None 
        self.tag_count: int = 0 
        self.tag_indexes: list[int] = [] 

        self.tags: list[Tag] = []
    
    def __repr__(self):
        return self.name
    
    def read(self, reader: Reader, version_4: bool =False) -> None:
        # Version 4 MSBTS have a short before the tag count that is the 
        # index of the tag group.
        if version_4:
            self.group_index = reader.read_uint16()

        self.tag_count = reader.read_uint16()
        self.tag_indexes = [reader.read_uint16() for _ in range(self.tag_count)]
        self.name = reader.read_string_nt()

    def write(self, writer: Writer) -> None:
        pass 
    
class Tag:
    """A class that represents a tag."""
    def __init__(self):
        self.name: str = ""
        self.parameter_count = 0 
        self.parameter_indexes: list[int] = []

        self.parameters: list[TagParameter] = []

    def __repr__(self):
        return self.name
    
    def read(self, reader: Reader) -> None:
        self.parameter_count = reader.read_uint16()
        self.parameter_indexes = [reader.read_uint16() for _ in range(self.parameter_count)]
        self.name = reader.read_string_nt()

    def write(self, writer: Writer) -> None:
        pass 

class TagParameter:
    """A class that represents a tag parameter."""
    def __init__(self):
        self.name: str = ""

        self.type: int = 0
        self.list_indexes: list[int] = []
        self.list_items: list[str] = []

    def __repr__(self):
        return self.name
    
    def read(self, reader: Reader) -> None:
        self.type = LMS_BinaryTypes(reader.read_uint8())

        if self.type is not LMS_BinaryTypes.LIST_INDEX:
            self.name = reader.read_string_nt() 
            return 
        
        reader.skip(1)
        list_count = reader.read_uint16()
        self.list_indexes = [reader.read_uint16() for _ in range(list_count)]
        self.name = reader.read_string_nt()

    def write(self, writer: Writer) -> None:
        pass 



    


from LMS.Common.LMS_Block import LMS_Block
from LMS.Common.LMS_Enum import LMS_BinaryTypes
from LMS.Project.MSBP import MSBP
from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer


class TSY1:
    """A class that represents a TSY1 block in a MSBT file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBT-File-Format#tsy1-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.style_indexes: list[int] = []

    def add_style_msbp(self, name: str, msbp: MSBP) -> None:
        """Adds the style index of a style from the msbp given a name.

        :param `name`: The name of the style.
        :param `msbp`: A MSBP object."""
        self.style_indexes.append(msbp.SLB1.get_index_by_label(name))

    def read(self, reader: Reader) -> None:
        """Reads the TSY1 block from a stream.

        :param `reader`: A Reader object."""
        self.block.read_header(reader)

        # There is no short for the amount of style indexes as its for every message
        # just read until the end is reached so the amount of messages doesnt need to be passed
        while reader.tell() != self.block.data_start + self.block.size:
            self.style_indexes.append(reader.read_uint32())

        self.block.seek_to_end(reader)

    def write(self, writer: Writer) -> None:
        """Writes the TSY1 block to a stream.

        :param `writer`: A Writer object."""
        self.block.write_initial_data(writer)

        for index in self.style_indexes:
            writer.write_uint32(index)

        self.block.size = 4 * len(self.style_indexes)
        self.block.write_end_data(writer)



from LMS.Common.LMS_Block import LMS_Block
from LMS.Common.LMS_Enum import LMS_BinaryTypes
from LMS.Project.MSBP import MSBP
from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer


class ATR1:
    """A class that represents a ATR1 block in a MSBT file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBT-File-Format#atr1-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        
        self.bytes_per_attribute = 0
        self.attributes: list[dict | bytes] = []
        self.structure: dict = {}

        # Define a string table attribute for when attributes arent decoded.
        # When attributes decoded a string table isn't needed. 
        self.strings: list[str] = []

    def attributes_decoded(self) -> bool:
        """Checks if the attributes are decoded and valid, as in they are not type bytes or are empty."""
        if self.attributes:
            return not isinstance(self.attributes[0], bytes)
        return False

    def read_encoded_attributes(self, reader: Reader) -> None:
        """Reads encoded attributes and the string table.

        :param `reader`: A Reader object."""
        self.attributes = [
            reader.read_bytes(self.bytes_per_attribute)
            for _ in range(self.attribute_count)
        ]

        # Verify ATR1 has a string table by verifying the size is bigger than the attributes
        if self.block.size > self.attribute_count * self.bytes_per_attribute:
            while reader.tell() != self.block.data_start + self.block.size:
                self.strings.append(reader.read_utf16_string())

    def read_decoded_attributes(self, reader: Reader) -> None:
        """Reads decoded attributes and the string table.

        :param `reader`: A Reader object."""
        for _ in range(self.attribute_count):
            attribute = {}
            for label in self.structure:
                attribute_type = self.structure[label].type

                if LMS_BinaryTypes._int_type(attribute_type):
                    attribute[label] = LMS_BinaryTypes.action_based_value(reader, attribute_type, None, "read")
                else:
                    if attribute_type is LMS_BinaryTypes.STRING:
                        offset = self.block.data_start + reader.read_uint32()
                        last_position = reader.tell()
                        reader.seek(offset)
                        utf16_string = reader.read_utf16_string()
                        attribute[label] = utf16_string
                        reader.seek(last_position)
                    else:
                        index = reader.read_uint8()
                        attribute[label] = self.structure[label].list_items[index]

            self.attributes.append(attribute)

    def get_bytes_per_attribute(self):
        """Gets the bytes_per_attribute of the ATR1 block."""
        if not self.attributes_decoded():
            return len(self.attributes[0])
        else:
            bytes_per_attribute = 0
            for label in self.structure:
                type = self.structure[label].type
                if LMS_BinaryTypes._8_bit_type(type):
                    bytes_per_attribute += 1
                elif LMS_BinaryTypes._16_bit_type(type):
                    bytes_per_attribute += 2
                elif (
                    LMS_BinaryTypes._32_bit_type(type) or type is LMS_BinaryTypes.STRING
                ):
                    bytes_per_attribute += 4
                else:
                    bytes_per_attribute += 1
            return bytes_per_attribute

    def write_encoded_attributes(self, writer: Writer) -> None:
        """Writes the encoded attributes to a stream.

        :param `writer`: a Writer object."""
        for attribute in self.attributes:
            writer.write_bytes(attribute)

        string_size = 8
        for string in self.strings:
            writer.write_utf16_string(string, use_double=True)
            string_size += len(string.encode(writer.get_utf16_encoding())) + 2
            
        self.block.size = self.attribute_count * self.bytes_per_attribute + string_size
        self.block.write_end_data(writer)

    def write_decoded_attributes(self, writer: Writer) -> None:
        """Writes the decoded attributes to a stream.

        :param `writer`: a Writer object."""
        self.strings = []

        string_offset = 8 + self.attribute_count * self.bytes_per_attribute

        for attribute in self.attributes:
            for label in self.structure:

                value = attribute[label]
                type = self.structure[label].type

                if LMS_BinaryTypes._int_type(type):
                    LMS_BinaryTypes.action_based_value(writer, type, value, "write")
                else:
                    if type is LMS_BinaryTypes.STRING:
                        self.strings.append(value)
                        writer.write_uint32(string_offset)
                        string_offset += (
                            len(value.encode(writer.get_utf16_encoding())) + 2
                        )
                    else:
                        writer.write_uint8(
                            self.structure[label].list_items.index(value)
                        )

    def create_attribute(self) -> dict:
        """Creates an attribute, decoded or encoded."""
        if self.structure:
            attribute = {}
            for label in self.structure:
                type = self.structure[label].type
                if LMS_BinaryTypes._get_bits(type) is not None:
                    attribute[label] = 0
                elif type is LMS_BinaryTypes.STRING:
                    attribute[label] = ""
                else:
                    attribute[label] = self.structure[label].list_items[0]
        else:
            attribute = b"\x00" * self.bytes_per_attribute
        return attribute

    def read(self, reader: Reader, msbp: MSBP = None) -> None:
        """Reads the ATR1 block from a stream.

        :param `reader`: a Reader object.
        :param `project`: a MSBP object used for decoding attributes."""

        self.block.read_header(reader)
        self.attribute_count = reader.read_uint32()
        self.bytes_per_attribute = reader.read_uint32()

        # Read the attributes
        if msbp is None or not msbp.ALB1.labels:
            self.read_encoded_attributes(reader)
            self.block.seek_to_end(reader)
            return

        self.structure = msbp.get_attribute_structure()
        self.read_decoded_attributes(reader)

        self.block.seek_to_end(reader)

    def write(self, writer: Writer) -> None:
        """Writes the ATR1 block to a stream.

        :param `writer`: a Writer object.
        """
        self.attribute_count = len(self.attributes)
        self.block.write_initial_data(writer, self.attribute_count)

        self.bytes_per_attribute = self.get_bytes_per_attribute()
        writer.write_uint32(self.bytes_per_attribute)

        if not self.attributes_decoded():
            self.write_encoded_attributes(writer)
            return

        self.write_decoded_attributes(writer)

        string_size = 8
        for string in self.strings:
            writer.write_utf16_string(string, use_double=True)
            string_size += len(string.encode(writer.get_utf16_encoding())) + 2

        self.block.size = self.attribute_count * self.bytes_per_attribute + string_size
        self.block.write_end_data(writer)



from enum import Enum, EnumMeta
from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer

class LMS_MessageEncoding(Enum):
    """An enum for representing message encoding. As in offical LMS files."""

    UTF8 = 0
    UTF16 = 1
    UTF32 = 2

class LMS_BinaryTypes(Enum):
    """A class that represents binary types utilized in attributes and tags."""

    UINT8_0 = 0
    UINT8_1 = 3
    FLOAT = 6
    UINT16_0 = 1
    UINT16_1 = 4
    UINT16_2 = 7
    UINT32_0 = 2
    UINT32_1 = 5
    STRING = 8
    LIST_INDEX = 9
    
    # Define a new str method to simplfy the string returned as the enum member name 
    def __str__(self) -> str:
        return self.name
    
    @classmethod
    def action_based_value(self, stream: Reader | Writer, value: "LMS_BinaryTypes", parameter = None, action: str = "read" or "write") -> None:
        bit_type = LMS_BinaryTypes._get_bits(value)
        if action == "read":
            if bit_type == "8":
                return stream.read_uint8()
            elif bit_type == "16":
                return stream.read_uint16()
            elif bit_type == "32":
                return stream.read_uint32()

        elif action == "write":
            if bit_type == "8":
                stream.write_uint8(parameter)
            elif bit_type == "16":
                stream.write_uint16(parameter)
            elif bit_type == "32":
                stream.write_uint32(parameter)

    @classmethod
    def _int_type(self, value: "LMS_BinaryTypes") -> bool:
        return LMS_BinaryTypes._get_bits(value) is not None
    
    @classmethod
    def _get_bits(self, value: "LMS_BinaryTypes") -> str:
        if self._8_bit_type(value):
            return "8"
        
        if self._16_bit_type(value):
            return "16"
        
        if self._32_bit_type(value):
            return "32"
        
        return None

    @classmethod
    def _8_bit_type(self, value: "LMS_BinaryTypes" or int):
        """Returns if a value or type is 8 bits.

        :param `value`: A int or LMS_Type enum value."""
        if isinstance(value, int):
            value = self(value)

        # Exlcude list type as it has special reading
        return value in (self.UINT8_0, self.UINT8_1)

    @classmethod
    def _16_bit_type(self, value: "LMS_BinaryTypes" or int):
        """Returns if a value or type is 16 bits.

        :param `value`: A int or LMS_Type enum value."""
        if isinstance(value, int):
            value = LMS_BinaryTypes(value)

        return value in (self.UINT16_0, self.UINT16_1, self.UINT16_2)

    @classmethod
    def _32_bit_type(self, value: "LMS_BinaryTypes" or int):
        """Returns if a value or type is 32 bits.

        :param `value`: A int or LMS_Type enum value."""
        if isinstance(value, int):
            value = self(value)

        # Exlcude String type as it has special reading
        return value in (self.UINT32_0, self.UINT32_1)



from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader


class CTI1:
    """A class that represents a CTI1 block in a MSBT file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBP-File-Format#cti1-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.source_files: list[str] = []

    def read(self, reader: Reader) -> None:
        """Reads the CTI1 block from a stream.

        :param `reader`: A Reader object."""
        self.block.read_header(reader)
        source_count = reader.read_uint32()

        # Read the sources
        for offset in self.block.get_item_offsets(reader, source_count):
            reader.seek(offset)
            source = reader.read_string_nt()
            self.source_files.append(source)



from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader
from LMS.Common.LMS_Enum import LMS_BinaryTypes
from LMS.Project.Structure import TagParameter


class TGP2:
    """A class that represents a TGP2 block in a MSBT file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBP-File-Format#tgp2-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.parameters: list[dict] = []

    def read(self, reader: Reader) -> None:
        """Reads the TGP2 block from a stream.

        :param `reader`: A Reader object."""
        self.block.read_header(reader)

        parameter_count = reader.read_uint16()
        reader.skip(2)
        # Read the parameters
        for offset in self.block.get_item_offsets(reader, parameter_count):
            parameter = TagParameter()
            reader.seek(offset)
            parameter.read(reader)
            self.parameters.append(parameter)

        self.block.seek_to_end(reader)



from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader
from LMS.Project.Structure import TagGroup

class ProjectBlock:
    """A class that represents all blocks in a MSBP file"""

    def __init__(self, item_type):
        self.block: LMS_Block = LMS_Block()
        self.item_type = item_type
        self.data: list[item_type] = []

    def read(self, reader: Reader, contains_offsets: bool = True, uint32_count: bool = True, version_4=False) -> None:
        """Reads the block from a stream.

        :param `reader`: a Reader object.
        :param `uint32_count`: determines weather the item count is a uint32. 
        :param `version_4`: determines if the block should be written for version 4 of a format. This is only used for `TagGroup`."""
        self.block.read_header(reader)

        if uint32_count:
            item_count = reader.read_uint32()
        else:
            item_count = reader.read_uint16()
            reader.skip(2)
        
        if contains_offsets:
            for offset in self.block.get_item_offsets(reader, item_count):
                reader.seek(offset)

                type = self.item_type()

                if isinstance(type, TagGroup):
                    type.read(reader, version_4)
                else:
                    type.read(reader)
                    
                self.data.append(type)
        else:
            for _ in range(item_count):
                type = self.item_type()
                type.read(reader)
                self.data.append(type)

        self.block.seek_to_end(reader)



from LMS.Project.MSBP import MSBP
from LMS.Common.LMS_Enum import LMS_BinaryTypes
from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer

from LMS.Message.Preset_Constants import (
    RESTRICTED_GLOBALS,
    ALLOWED_ATTRIBUTES,
    SYSTEM_STRUCTURE,
)

from lupa.lua54 import LuaRuntime

import importlib.resources
import json


def filter_attribute_access(obj, attr_name, is_setting):
    if isinstance(attr_name, str) and attr_name not in ALLOWED_ATTRIBUTES:
        raise AttributeError("You cannot access this attribute.")
    return attr_name


class Preset:
    """A static class for loading and storing a preset."""

    def __init__(self, global_list: list[str] = None):
        self.lua_runtime = LuaRuntime(attribute_filter=filter_attribute_access)
        self.structure: list[dict] = [SYSTEM_STRUCTURE]

        # Load the base Preset
        with importlib.resources.open_text("LMS.Message", "System.lua") as file:
            self.base_preset = dict(self.lua_runtime.execute(file.read()))
            self.stream_functions = self.base_preset

        globals = self.lua_runtime.globals()

        restricted_globals = global_list if global_list else RESTRICTED_GLOBALS
        for name in restricted_globals:
            globals[name] = None

    def get_indexes_by_name(self, group_name: str, tag_name: str) -> tuple:
        """Get the indexes of the group and tag in the Preset.

        :param `group_name`: the name of the group.
        :param `tag_name`: the name of the tag."""
        group_index = None
        tag_index = None

        for i in self.stream_functions:
            if self.stream_functions[i].name == group_name:
                group_index = i

        for i in self.stream_functions[group_index].tags:
            if self.stream_functions[group_index].tags[i].name == tag_name:
                tag_index = i

        return group_index, tag_index

    def get_function_data(self, group: int | str, tag: int | str) -> tuple:
        """Get the table of data in a function.

        :param `group`: the name/index of the group.
        :param `tag`: the name/index of the tag."""
        if isinstance(group, str) and isinstance(tag, str):
            group, tag = self.get_indexes_by_name(group, tag)

        stream_function = self.stream_functions[group].tags[tag].stream_function

        if stream_function is None:
            return None, None, None

        result = stream_function()
        return result[1], result[2], result[3]

    def load_preset_file(self, preset_path: str) -> None:
        """Loads a preset file.

        :param `preset_path`: the path of the .lua Preset.
        """
        self.stream_functions = self.base_preset

        with open(preset_path, "r") as file:
            self.stream_functions.update(dict(self.lua_runtime.execute(file.read())))

    @staticmethod
    def create_preset_file(preset_path: str, msbp: MSBP | None = None) -> None:
        """Creates a lua Preset.

        :param `preset_path`: the path to generate the Lua preset.
        :param `msbp`: a project object for a specifc game.
        """
        """Creates a preset file."""
        if msbp is None:
            structure = []
        else:
            structure = msbp.get_tag_structure()

        function_template = """
local function {name}()
{parameter_info}
    local function read(data, reader)
{read_body}    end
    local function write(data, writer)
{write_body}    end 
    return {{read, write, parameter_info}}
end
"""
        # Index function for writing the index of list ltems
        index_function = "local function index(table, value) for i, v in ipairs(table) do if v == value then return i - 1 end end end\n"

        lua_code = ["local stream_functions = {}\n\n", index_function]

        # Slice by 1 to avoid the System group
        for group in structure[1:]:
            for tag in group.tags:
                parameter_info, read_body, write_body = (
                    Preset._generate_function_bodies(tag)
                )

                lua_code.append(
                    function_template.format(
                        name=f"{group.name.lower()}_{tag.name.lower()}",
                        parameter_info=parameter_info,
                        read_body=read_body,
                        write_body=write_body,
                    )
                )

        # Build the final stream_functions table
        lua_code.append("\nstream_functions = {\n")

        for i, group in enumerate(structure[1:], start=1):
            lua_code.append(
                f"\t[{i}] = {{\n\t\tname = '{group.name}',\n\t\ttags = {{\n"
            )

            for i, tag in enumerate(group.tags):
                lua_code.append(
                    f"\t\t\t[{i}] = {{name = '{tag.name}', stream_function = {group.name.lower()}_{tag.name.lower()}}},\n"
                )

            lua_code.append("\t\t}\n\t},\n")

        lua_code.append("}\n\nreturn stream_functions")
        lua_code.insert(
            0,
            """--[[ 
   This is a preset used for parsing tags. 
   DO NOT edit any variable which are list items (tables that appear at the top of a tag function). It will lead to errors.
   Make sure you peserve parameter names when editing a function.
--]]\n\n""",
        )

        with open(preset_path, "w+") as file:
            file.truncate()
            file.write("".join(lua_code))

    def _generate_function_bodies(tag):
        info_body = ["\tlocal parameter_info = {\n"]
        read_body = []
        write_body = []

        body_template = {
            LMS_BinaryTypes.FLOAT: (
                "\t\tdata['{parameter_name}'] = reader.read_float()\n",
                "\t\twriter.write_float(tonumber(data['{parameter_name}']))\n",
            ),
            LMS_BinaryTypes.LIST_INDEX: (
                "\t\tdata['{parameter_name}'] = parameter_info['{parameter_name}'].list_items[reader.read_uint8({{lua_index=true}})]\n",
                "\t\twriter.write_uint8(index(parameter_info['{parameter_name}'].list_items, data['{parameter_name}']))\n",
            ),
            LMS_BinaryTypes.STRING: (
                "\t\tdata['{parameter_name}'] = reader.read_len_prefixed_utf16_string()\n",
                "\t\twriter.write_len_prefixed_utf16_string(data['{parameter_name}'])\n",
            ),
        }

        data_table = "\t\t{parameter_name} = {{type = {parameter_type}, list_items = {list_items}}},\n"

        for parameter in tag.parameters:
            info_body.append(
                data_table.format(
                    parameter_name=parameter.name,
                    parameter_type=parameter.type.value,
                    list_items=(
                        set(parameter.list_items) if parameter.list_items else "{}"
                    ),
                )
            )

            if parameter.type not in body_template:
                read_body.append(
                    f"\t\tdata['{parameter.name}'] = reader.read_uint{LMS_BinaryTypes._get_bits(parameter.type)}()\n"
                )
                write_body.append(
                    f"\t\twriter.write_uint{LMS_BinaryTypes._get_bits(parameter.type)}(tonumber(data['{parameter.name}']))\n"
                )
            else:
                read, write = body_template[parameter.type]
                read_body.append(read.format(parameter_name=parameter.name))
                write_body.append(write.format(parameter_name=parameter.name))

        info_body.append("}")
        return "".join(info_body), "".join(read_body), "".join(write_body)



ALLOWED_ATTRIBUTES = [
    "get_utf16_encoding",
    "change_byte_order",
    "skip",
    "read_bytes",
    "seek",
    "tell",
    "read_uint8",
    "read_uint16",
    "read_float",
    "read_uint32",
    "read_string_len",
    "read_string_nt",
    "read_utf16_string",
    "read_len_prefixed_utf16_string",
    "write_uint8",
    "write_uint16",
    "write_uint32",
    "write_float",
    "write_len_prefixed_utf16_string",
    "write_utf16_string",
    "write_bytes",
    "write_string_nt",
    "write_string",
]

RESTRICTED_GLOBALS = [
        "python", "io", "os", "debug", "package", "dofile",
        "load", "loadfile", "require", "collectgarbage",
        "rawget", "rawset", "getmetatable", "setmetatable", "coroutine"
]

SYSTEM_STRUCTURE = {
    "name": "System",
    "tags": [
      {
        "name": "Ruby",
        "parameters": [
          {
            "name": "rt",
            "type": 8,
            "list_items": []
          }
        ]
      },
      {
        "name": "Font",
        "parameters": [
          {
            "name": "face",
            "type": 8,
            "list_items": []
          }
        ]
      },
      {
        "name": "Size",
        "parameters": [
          {
            "name": "percent",
            "type": 1,
            "list_items": []
          }
        ]
      },
      {
        "name": "Color",
        "parameters": [
          {
            "name": "r",
            "type": 0,
            "list_items": []
          },
          {
            "name": "g",
            "type": 0,
            "list_items": []
          },
          {
            "name": "b",
            "type": 0,
            "list_items": []
          },
          {
            "name": "a",
            "type": 0,
            "list_items": []
          },
        ]
      },
      {
        "name": "PageBreak",
        "parameters": []
      }
    ]
}


from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer

from LMS.Common.LMS_Enum import LMS_MessageEncoding

class LMS_Binary:
    """A class that represents common data and functions shared betweeen all LMS files."""

    def __init__(self):
        self.magic: str = None
        self.bom: str = None
        self.encoding: LMS_MessageEncoding = LMS_MessageEncoding.UTF8
        self.revision: int = None
        self.block_count: int = None
        self.file_size: int = None

    def read_header(self, reader: Reader) -> None:
        """Reads the header from the stream.

        :param `reader`: a Reader object."""
        self.magic = reader.read_string_len(8)
        self.bom = "little" if reader.read_bytes(2) == b"\xFF\xFE" else "big"

        reader.byte_order = self.bom
        # Unk 1
        reader.skip(2)
        self.encoding = LMS_MessageEncoding(reader.read_uint8())

        self.revision = reader.read_uint8()
        self.block_count = reader.read_uint16()
        # Unk 2
        reader.skip(2)
        self.file_size = reader.read_uint32()
        # Skip to end
        reader.skip(10)
        end = reader.tell()

        # Verify file size alignment
        reader.seek(0, 2)
        size = reader.tell()
        if size != self.file_size:
            raise Exception("The file size is misaligned!")

        reader.seek(end)
    
    def write_header(self, writer: Writer) -> None:
        """Writes the header from the stream.

        :param `writer`: A Writer object."""
        self.bom = writer.byte_order
        writer.write_string(self.magic)
        writer.write_bytes(b"\xFF\xFE" if self.bom ==
                           "little" else b"\xFE\xFF")
        # Unk 1
        writer.write_bytes(b"\x00\x00")

        writer.write_uint8(self.encoding.value)
        writer.write_uint8(self.revision)
        writer.write_uint16(self.block_count)
        # Unk 2
        writer.write_bytes(b"\x00\x00")
        writer.write_uint32(0)
        # Padding
        writer.write_bytes(b"\x00" * 10)

    def write_size(self, writer: Writer):
        """Writes the size of the file to the stream.

        :param `writer`: A Writer object."""
        writer.seek(0, 2)
        size = writer.tell()
        writer.seek(18)
        writer.write_uint32(size)

    def search_block_by_name(self, reader: Reader, name: str) -> tuple[bool, int]:
        """Returns the absolute offset of a specific block from a stream and if the offset is valid.

        :param `reader`: A Reader object.
        :param `name`: The name of the block."""
        name = name.upper()
        blocks = self.search_all_blocks(reader)
        return (True, blocks[name]) if name in blocks else (False, None)

    def search_all_blocks(self, reader: Reader) -> dict[str:int]:
        """Returns the absolute offset of every block in a stream.

        :param `reader`: A Reader object."""
        result = {}
        # First block is always located absolute 32
        reader.seek(32)
        block_count = self.block_count

        for _ in range(block_count):
            offset = reader.tell()
            magic = reader.read_string_len(4)
            size = reader.read_uint32()
            # Skip Padding to seek past block using size
            reader.skip(8)
            reader.seek(size, 1)
            remainder = 16 - size % 16

            if remainder != 16:
                reader.skip(remainder)

            result[magic] = offset

        return result



import struct
from io import BytesIO


types = {
    "little": {"uint8": "<B", "uint16": "<H", "uint32": "<I", "float": "<f"},
    "big": {"uint8": ">B", "uint16": ">H", "uint32": ">I", "float": ">f"},
}


class Writer:
    """A class utilized for writing to a stream."""

    def __init__(self, data: BytesIO, byte_order: str = "little"):
        self.data = BytesIO(data)
        self.byte_order = byte_order

    def get_utf16_encoding(self) -> str:
        """Returns the UTF-16 encoding of the stream."""
        return "UTF-16-LE" if self.byte_order == "little" else "UTF-16-BE"

    def skip(self, length: int) -> None:
        """Skips `length` amount of bytes."""
        self.data.read(length)

    def write_bytes(self, data: bytes | int) -> None:
        """Writes specifc bytes to the stream."""
        if isinstance(data, int):
            data = data.to_bytes(1, "little")

        self.data.write(data)

    def seek(self, offset: int, whence: int = 0) -> None:
        """Seeks to an offset with whence."""
        self.data.seek(offset, whence)

    def tell(self) -> int:
        """Returns the current position in the stream."""
        return self.data.tell()

    def write_string_nt(self, string: str) -> None:
        """Writes a null terminated string to the stream."""
        self.data.write(string.encode("UTF-8") + b"\x00")

    def write_string(self, string: str) -> None:
        """Writes a  string to the stream."""
        self.data.write(string.encode("UTF-8"))

    def write_uint8(self, num: int):
        """Writes a UInt8 to the stream."""
        self.data.write(struct.pack(types[self.byte_order]["uint8"], num))

    def write_uint16(self, num: int):
        """Writes a UInt16 to the stream."""
        self.data.write(struct.pack(types[self.byte_order]["uint16"], num))

    def write_uint32(self, num: int):
        """Writes a UInt32 to the stream."""
        self.data.write(struct.pack(types[self.byte_order]["uint32"], num))

    def write_float(self, num: float):
        """Writes a Float to the stream."""
        self.data.write(struct.pack(types[self.byte_order]["float"], num))
        
    def write_utf16_string(self, string: str, use_double: bool = False):
        """Writes a  UTF-16 string to a stream."""
        encoding = "UTF-16-LE" if self.byte_order == "little" else "UTF-16-BE"
        self.write_bytes(string.encode(encoding))
        if use_double:
            self.write_bytes(b"\x00\x00")

    def write_len_prefixed_utf16_string(self, string: str):
        """Writes a length prefixed UTF-16 string to a stream."""
       
        encoding = "UTF-16-LE" if self.byte_order == "little" else "UTF-16-BE"
        string = string.encode(encoding)
  
        self.write_uint16(len(string))

        self.write_bytes(string)

    def get_data(self) -> bytes:
        return self.data.getvalue()



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="PylibMS",
    version="0.7.4",
    author="AbdyyEee",
    packages=find_packages(),
    description="Library for Nintendo's LMS file formats (MSBT, MSBP) for Nintendo 3DS and Wii U and Switch.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AbdyyEee/PylibMS",
    keywords=["nintendo", "msbf", "msbt", "msbp",
              "modding", "lms", "libmessagestudio"],
    python_requires=">=3.10",
    include_package_data=True,
    package_data={
        'LMS.Message': ['System.lua'],
    },
    install_requires=[
        "lupa>=2.2"
    ]
)



import re
import struct
import logging

from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer
from LMS.Project.MSBP import MSBP
from LMS.Message.Preset import Preset

system_names = {0: "Ruby", 1: "Font", 2: "Size", 3: "Color", 4: "PageBreak"}


class Tag_Utility:
    """Static class used to house most tag related functions."""

    @staticmethod
    def contains_tag(message: str, group_name: str, tag_name: str) -> bool:
        """Returns if a message contains a specifc tag.

        :param `message`: the message to search.
        :param `group_name`: the group name of the tag.
        :param`tag_name`: the name of the tag."""
        for part in Tag_Utility.split_message_by_tag(message):
            if not Tag_Utility.tag_encoded(part):
                data = Tag_Utility.get_decoded_tag_information(part)
                if data["group_name"] == group_name and data["tag_name"] == tag_name:
                    return True

        return False

    @staticmethod
    def create_tag(
        group_name: str, tag_name: str, parameters: str | dict | None = None
    ) -> str:
        if isinstance(parameters, dict):
            parameters = Tag_Utility.get_str_parameter_representation(parameters)

        if parameters:
            return f"<{group_name}::{tag_name} {parameters}>"

        return f"<{group_name}::{tag_name}>"

    @staticmethod
    def get_group_and_tag_names(tag: str) -> tuple[str, str]:
        """Returns a list of the group and tag name given a tag.

        :param `tag`: The tag to get the information for."""
        result = re.search(r"<([^:]+)::([^ >]+)[^>]*>", tag)
        return (result.group(1), result.group(2))

    @staticmethod
    def has_parameters(tag: str) -> bool:
        """Returns if a tag has parameters or not.

        :param `tag`: The tag to check."""
        return True if Tag_Utility.get_dict_parameter_representation(tag) else False

    @staticmethod
    def is_tag(message: str) -> bool:
        """Returns if a message is a tag.

        :param `message`: The message to check."""
        return (
            not message.startswith("<\\")
            and message.startswith("<")
            and message.endswith(">")
        )

    @staticmethod
    def split_message_by_tag(message: str) -> tuple[str]:
        """Splits a message by the control tags.

        :param `message`: The message to split."""
        return tuple(re.split(r"(<[^>]+>)", message))

    @staticmethod
    def tag_encoded(tag: str) -> bool:
        """Returns if a tag is encoded or not.

        :param `tag`: The tag to check."""
        return "." in tag[: tag.find(":")]

    @staticmethod
    def read_tag(reader: Reader, preset: Preset) -> tuple:
        """Reads a tag from a stream, encoded or decoded.

        :param `reader`: a Reader object.
        :param `preset`: a Preset object.
        :param `msbp`: a MSBP object"""
        group_index = reader.read_uint16()
        tag_index = reader.read_uint16()
        
        # When the structure is a length of 1 it means that a preset was never loaded and it
        # contains just the 'System' functions. Using this check prevents reading from it
        if len(preset.stream_functions) == 1 and group_index:
            return Tag_Utility.read_encoded_tag(reader, group_index, tag_index), None

        return Tag_Utility.read_decoded_tag(reader, preset, group_index, tag_index)

    @staticmethod
    def read_encoded_tag(reader: Reader, group_index: int, tag_index: int) -> str:
        """Reads an encoded tag from a stream.

        :param `reader`: a Reader object.
        :param `group_index`: the index of the tag group.
        :param `tag_index`: the index of the tag in the group.
        """
        parameter_size = reader.read_uint16()
        hex_parameters = reader.read_bytes(parameter_size).hex()
        encoded_parameters = "-".join(
            [
                hex_parameters[i] + hex_parameters[i + 1]
                for i in range(0, len(hex_parameters), 2)
            ]
        )
        if parameter_size == 1:
            parameter_size = 0

        return f"<n{group_index}.{tag_index}:{encoded_parameters}>"

    @staticmethod
    def read_decoded_tag(
        reader: Reader, preset: Preset, group_index: int, tag_index: int
    ) -> tuple:
        """Reads an encoded tag from a stream.

        :param `reader`: a Reader object.
        :param `preset`: a Preset object.
        :param `msbp`: a MSBP object
        :param `group_index`: the index of the tag group.
        :param `tag_index`: the index of the tag in the group.
        """
        start = reader.tell()
        parameters = {}

        read, _, parameter_info = preset.get_function_data(group_index, tag_index)

        if read is None:
            reader.seek(start)
            return Tag_Utility.read_encoded_tag(reader, group_index, tag_index), None

        group_name = preset.stream_functions[group_index].name
        tag_name = preset.stream_functions[group_index].tags[tag_index].name
        function_name = f"{group_name.lower()}_{tag_name.lower()}"

        parameter_size = reader.read_uint16()
        end = reader.tell() + parameter_size

        try:
            read(parameters, reader)
        except Exception as error:
            parameter = list(parameter_info.keys())[len(parameters)]

            error = f"An error occurred in the function '{function_name}', reading the parameter '{parameter}' at start offset {start}. {error}"
            reader.seek(start)
            return Tag_Utility.read_encoded_tag(reader, group_index, tag_index), error

        # Account for 0xCD padding
        reader.seek(end + 1 if parameter_size % 2 == 1 else end)
        return Tag_Utility.create_tag(group_name, tag_name, parameters), None

    @staticmethod
    def write_tag(writer: Writer, tag: str, preset: Preset) -> None | str:
        """Writes both encoded and decoded tags to a stream.

        :param `writer`: a Writer object.
        :param `tag`: the tag to write.
        :param `preset`: a Preset object.
        :param `msbp`: a MSBP object."""

        if Tag_Utility.tag_encoded(tag):
            return Tag_Utility.write_encoded_tag(writer, tag)

        return Tag_Utility.write_decoded_tag(writer, tag, preset)

    @staticmethod
    def write_encoded_tag(writer: Writer, tag: str) -> None:
        """Writes an encoded tag to a stream.

        :param `writer`: a Writer object.
        :param `tag`: the tag to write."""

        group_index = int(tag[2 : tag.index(".")])
        tag_index = int(tag[tag.index(".") + 1 : tag.index(":")])

        parameters = tag[tag.rindex(":") + 1 : len(tag) - 1].split("-")
        parameter_size = len(parameters)

        if parameter_size == 1:
            parameter_size = 0

        writer.write_uint16(int(group_index))
        writer.write_uint16(int(tag_index))
        writer.write_uint16(parameter_size)

        for parameter in parameters:
            if parameter:
                writer.write_bytes(bytes.fromhex(parameter))

    @staticmethod
    # It is simple to handle errors by returning a string when reading decoded tags
    def write_decoded_tag(writer: Writer, tag: str, preset: Preset) -> str | None:
        """Writes a decoded tag to a stream.

        :param `writer`: a Writer object.
        :param `tag`: the tag to write.
        :param `msbp`: a MSBP object.
        :param `preset`: a Preset object."""
        tag_info = Tag_Utility.get_decoded_tag_information(tag)

        group_index, tag_index = preset.get_indexes_by_name(
            tag_info["group_name"], tag_info["tag_name"]
        )
        write_function = preset.get_function_data(group_index, tag_index)[1]

        group_name = preset.stream_functions[group_index].name
        tag_name = preset.stream_functions[group_index].tags[tag_index].name
        function_name = f"{group_name}_{tag_name}"

        writer.write_uint16(group_index)
        writer.write_uint16(tag_index)
        size_offset = writer.tell()
        writer.write_uint16(0)
        start = writer.tell()

        try:
            write_function(tag_info["parameters"], writer)
        except Exception as exception:
            return f"An error occurred while writing the tag in the function {function_name} at start offset {start}. {exception}"

        size = writer.tell() - start

        if size % 2 == 1:
            size += 1
            writer.write_bytes(b"\xCD")

        writer.seek(size_offset)
        writer.write_uint16(size)
        writer.seek(size, 1)

    @staticmethod
    def get_decoded_tag_information(tag: str) -> dict:
        """Returns all the information of a decoded tag.

        :param `tag`: The tag to get the information for."""
        group_name, tag_name = Tag_Utility.get_group_and_tag_names(tag)

        if not Tag_Utility.has_parameters(tag):
            return {"group_name": group_name, "tag_name": tag_name, "parameters": {}}

        parameters = Tag_Utility.get_dict_parameter_representation(tag)
        return {
            "group_name": group_name,
            "tag_name": tag_name,
            "parameters": parameters,
        }

    @staticmethod
    def get_str_parameter_representation(parameters: dict) -> str:
        """Returns the parameters in a dictionary as the formatted part of a tag.

        :param `parameters`: The parameters with key being the name and the values being what is enclosed.
        """
        result = ""
        for parameter in parameters:
            result += f'{parameter}="{parameters[parameter]}" '
        return result.strip()

    @staticmethod
    def get_dict_parameter_representation(tag: str) -> dict:
        """Returns the parameters in a dictionary.

        :param `tag`: The tag to get the parameters for."""
        return dict(re.findall(r'(\w+)="([^"]*)"', tag))



from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader
from LMS.Project.Structure import Tag

class TAG2:
    """A class that represents a TAG2 block in a MSBT file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBP-File-Format#tag2-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.tags: list[dict] = []

    def read(self, reader: Reader) -> None:
        """Reads the TAG2 block from a stream.

        :param `reader`: A Reader object."""
        self.block.read_header(reader)
        tag_count = reader.read_uint16()
        reader.skip(2)
   
        for offset in self.block.get_item_offsets(reader, tag_count):
            tag = Tag()
            reader.seek(offset)
            tag.read(reader)
            self.tags.append(tag)

        self.block.seek_to_end(reader)



from LMS.Common.LMS_Binary import LMS_Binary
from LMS.Common.LMS_Enum import LMS_MessageEncoding
from LMS.Common.LMS_Enum import LMS_BinaryTypes
from LMS.Project.MSBP import MSBP

from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer

from LMS.Common.LMS_HashTable import LMS_HashTable
from LMS.Message.TXT2 import TXT2
from LMS.Message.ATR1 import ATR1
from LMS.Message.TSY1 import TSY1

from LMS.Message.Preset import Preset


class MSBT:
    """A class that represents a Message Studio Binary Text file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBT-File-Format"""

    def __init__(self):
        self.binary: LMS_Binary = LMS_Binary()
        self.LBL1: LMS_HashTable = LMS_HashTable()
        self.ATR1: ATR1 = ATR1()
        self.TSY1: TSY1 = TSY1()
        self.TXT2: TXT2 = TXT2()

        # Set the magic for each block
        self.LBL1.block.magic = "LBL1"
        self.ATR1.block.magic = "ATR1"
        self.TSY1.block.magic = "TSY1"
        self.TXT2.block.magic = "TXT2"

    def add_data(self, label: str, ignore_atr1=False, ignore_tsy1=False) -> None:
        """Adds a new instance of every type to each block.

        :param `label`: the label for the LBL1 block.
        :param `ignore_atr1`: bool to add an attribute to ATR1 or to ignore it.
        :param `ignore_tsy1`: bool to add an attribute to TSY1 or to ignore it."""
        self.LBL1.add_label(label)

        # Add the default values for each block
        self.TXT2.messages.append("")

        if self.ATR1 is not None and not ignore_atr1:
            attribute = self.ATR1.create_attribute()
            self.ATR1.attributes.append(attribute)

        if self.TSY1 is not None and not ignore_tsy1:
            self.TSY1.style_indexes.append(0)

    def combine_messages(self) -> dict:
        """Returns a dictionary of every label mapped to a message."""
        return {self.LBL1.labels[i]: self.TXT2.messages[i] for i in self.LBL1.labels}

    def combine_attributes(self) -> dict:
        """Returns a dictionary of every label mapped to an attribute."""
        return {self.LBL1.labels[i]: self.ATR1.attributes[i] for i in self.LBL1.labels}

    def combine_styles(self, msbp: MSBP) -> dict:
        """Returns a dictionary of every label mapped to a style.

        :param `msbp`: A MSBP object."""
        return {
            self.LBL1.labels[i]: msbp.SLB1.labels[self.TSY1.style_indexes[i]]
            for i in self.LBL1.labels
        }

    def read(self, reader: Reader, preset: Preset = None, msbp: MSBP = None):
        """Reads a MSBT file from a stream.

        :param `reader`: A Reader object.
        :param `msbp`: A MSBP object. Used for decoding of attributes and tags
        """
        if preset is None:
            preset = Preset()


        self.binary.read_header(reader)

        lbl1_valid, lbl1_offset = self.binary.search_block_by_name(reader, "LBL1")
        atr1_valid, atr1_offset = self.binary.search_block_by_name(reader, "ATR1")
        tsy1_valid, tsy1_offset = self.binary.search_block_by_name(reader, "TSY1")
        txt2_valid, txt2_offset = self.binary.search_block_by_name(reader, "TXT2")

        # Read LBL1
        if lbl1_valid:
            reader.seek(lbl1_offset)
            self.LBL1.read(reader)
        else:
            self.LBL1 = None

        # Add all the labels to the TXT2 errors dictionary 
        # the errors are set in TXT2
        for i in self.LBL1.labels:
            self.TXT2.errors[self.LBL1.labels[i]] = {}
            self.TXT2.errors[self.LBL1.labels[i]]["read"] = []
            self.TXT2.errors[self.LBL1.labels[i]]["write"] = []

        if atr1_valid:
            reader.seek(atr1_offset)
            self.ATR1.read(reader, msbp)
        else:
            self.ATR1 = None

        # Read TSY1
        if tsy1_valid:
            reader.seek(tsy1_offset)
            self.TSY1.read(reader)
        else:
            self.TSY1 = None
        # Read TXT2
        if txt2_valid:
            reader.seek(txt2_offset)
            self.TXT2.read(reader, preset, self.LBL1)
        else:
            self.TXT2 = None

    def write(self, writer: Writer, preset: Preset = None) -> None:
        """Writes a MSBT file to a stream.

        :param `reader`: A Reader object."""
        if preset is None:
            preset = Preset()

        writer.byte_order = self.binary.bom

        block_count = 0

        if self.LBL1 is not None:
            block_count += 1

        if self.ATR1 is not None:
            block_count += 1

        if self.TSY1 is not None:
            block_count += 1

        if self.TXT2 is not None:
            block_count += 1

        self.binary.magic = "MsgStdBn"
        
        if not self.binary.encoding:
            self.binary.encoding = LMS_MessageEncoding(1)

        self.binary.revision = 3
        self.binary.block_count = block_count

        self.binary.write_header(writer)

        if self.LBL1 is not None:
            self.LBL1.write(writer, 101)

        if self.ATR1 is not None:
            self.ATR1.write(writer)

        if self.TSY1 is not None:
            self.TSY1.write(writer)

        if self.TXT2 is not None:
            self.TXT2.write(writer, preset, self.LBL1)

        writer.seek(0, 2)
        size = writer.tell()
        writer.seek(18)
        writer.write_uint32(size)
